/*
 * Simple mount interface for userfs.  This does permissions checking
 * and allows non-suid processes to be mounted.
 *
 * It checks the proposed mount dir, does the mount and starts up
 * the filesystem program attached to the kernel with pipes to
 * deal with the filesystem operations.  The filesystem process
 * runs as the invoking user.
 *
 * Muserfs expects the child to only ever exit properly when it
 * gets an EOF from the kernel.  Muserfs takes signals and umounts
 * the filesystem when it gets normal termination signals.
 * muserfs will also attempt to unmount the filesystem if the child
 * exists unexpectedly.  If this happens, muserfs will exit regardless
 * of whether it was successful.
 * 
 * The /etc/mtab file is also updated correctly, so that umount should
 * be able to unmount the filesystem.
 *
 * The input and output fds are inhereted by the filesystem process
 * and are passed on the command-line with the -i fd (kernel to process)
 * and -o fd (process to kernel) options.
 *
 * Usage: muserfs [-a] program mount-point [args...]
 *
 * This program is distributed under the terms of the
 * Free Software Foundation General Public Licence.
 * Copyright Jeremy Fitzhardinge 1993
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <mntent.h>
#include <pwd.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/stat.h>

#include <linux/fs.h>
#include <linux/userfs_fs.h>
#include <linux/userfs_mount.h>
#include <linux/userfs_types.h>
#include <linux/userfs_fs_sb.h>

char *progname = NULL;

const char *mount_point = NULL;

int doumount(const char *);
int mounted = 0;

void sighandler(int sig)
{
	switch (sig)
	{
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
	case SIGQUIT:
	case SIGCHLD:
		if (!mounted)
			break;
		
		if (doumount(mount_point))
			exit(0);
		else
			fprintf(stderr, "%s: failed to unmount \"%s\"\n",
				progname, mount_point);

	default:
		return;
	}
}

void sysfatal(const char *str, ...)
{
	va_list va;
	int er = errno;
	
	va_start(va, str);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, str, va);
	fprintf(stderr, ": %s\n", strerror(er));
	va_end(va);
	exit(1);
}

void *xmalloc(size_t s)
{
	void *m = malloc(s);

	if (m == NULL)
		sysfatal("memory allocation failure");

	return m;
}

char *fullpath(const char *p)
{
	char cwd[1024];
	int dl;
	char *np;
	
	if (*p == '/')
		return strdup(p);

	if (getcwd(cwd, sizeof(cwd)) == NULL)
		sysfatal("Can't get cwd");

	dl = strlen(cwd);
	np = xmalloc(dl + strlen(p) + 2);

	strcpy(np, cwd);
	np[dl] = '/';
	strcpy(&np[dl]+1, p);

	return np;
}

char *makeopts(char **av, int narg, const char *name, int pid)
{
	int i, j, nblen;
	char *op, *cp;
	char nbuf[100];
	
	sprintf(nbuf, "user=%s,pid=%d,nodev,nosuid,", name, pid);
	nblen = j = strlen(nbuf);
	
	for(i = 0; i < narg; i++)
		j += strlen(av[i]) + 1;

	op = xmalloc(j + 1);

	strcpy(op, nbuf);
	cp = &op[nblen];
	
	for(i = 0; i < narg; i++)
	{
		strcpy(cp, av[i]);
		cp += strlen(av[i]);
		*cp++ = ',';
	}

	cp[-1] = 0;

	return op;
}

int doumount(const char *mpoint)
{
	int kid;
	int wstat;
	int i;
	static volatile int interlock;
	static const char *umounts[] =
	{
		"/sbin/umount",
		"/etc/umount",
		"/usr/sbin/umount",
		"/bin/umount",
	};
	
	if (++interlock != 1)
		return 0;
	if (mpoint == NULL)
		return 0;
	
	switch(kid = fork())
	{
	case -1:
		perror("umount fork failed");
		return 1;
	case 0:
		setuid(geteuid());
		for(i = 0; i < sizeof(umounts)/sizeof(*umount); i++)
			execl(umounts[i], "umount", mpoint, NULL);
		sysfatal("exec of \"umount %s\" failed", mpoint);
		exit(1);
	}

	if (wait4(kid, &wstat, 0, NULL) == -1)
		sysfatal("Wait for umount failed");

	interlock--;
	
	if (WEXITSTATUS(wstat) != 0)
		return 0;
	mounted = 0;
	return 1;
}

/*
 * Check filesystem process is there and functioning by asking if
 * it supports the upp_mount operation.  If it doesn't or fails to
 * respond then the mount is aborted.
 */
int checkfs(int in, int out)
{
	up_preamble pre;
	upp_repl repl;
	unsigned char buf[40];
	int sz;
	
	pre.version = UP_VERSION;
	pre.op = up_mount;
	pre.isreq = UP_ENQ;
	pre.seq = 0;
	pre.size = 0;

	assert(sizeof(buf) >= sizeof_up_preamble(&pre));
	assert(sizeof(buf) >= sizeof_upp_repl(&repl));

	encode_up_preamble(&pre, buf);
	sz = sizeof_up_preamble(&pre);
	
	if (write(out, buf, sz) != sz)
		return 0;

	sz = sizeof_upp_repl(&repl);
	
	if (read(in, buf, sz) != sz)
		return 0;

	decode_upp_repl(&repl, buf);

	if (repl.version != UP_VERSION || repl.op != up_mount ||
	    repl.seq != 0 || repl.isreq != UP_REPL ||
	    repl.size != 0 || repl.errno != 0)
		return 0;
	
	return 1;
}

void domount(int infd, int outfd, const char *mpoint, char *prog,
	     char **args, int nargs, int kid)
{
	struct userfs_mount mnt;
	struct mntent ment;
	FILE *mtab;
	struct passwd *pwd;
	char *mdev;
	int fd;
	
	if ((pwd = getpwuid(getuid())) == NULL)
		sysfatal("can't get password file entry for uid %d",
			 getuid());
	
	mnt.tokern = infd;
	mnt.fromkern = outfd;
	mnt.version = USERFS_VERSION;
	mnt.magic = USERFS_SUPER_MAGIC;
	
	if (mount(NULL, mpoint, "userfs",
		  MS_MGC_VAL|MS_NODEV|MS_NOSUID, &mnt) == -1)
		sysfatal("mount of %s on %s failed",
			 prog, mpoint);

	close(infd);
	close(outfd);

	mdev = xmalloc(strlen(prog) + 7);
	sprintf(mdev, "%s:%d", prog, kid);
	
	ment.mnt_fsname = mdev;
	ment.mnt_dir = fullpath(mpoint);
	ment.mnt_type = "userfs";
	ment.mnt_opts = makeopts(args, nargs, pwd->pw_name, kid);
	ment.mnt_freq = 0;
	ment.mnt_passno= 0;

	mount_point = ment.mnt_dir;
	
	if ((fd = open(MOUNTED"~", O_RDWR|O_CREAT|O_EXCL, 0600)) == -1)
		sysfatal("Can't get "MOUNTED"~ lock file");
	close(fd);
	
	if ((mtab = setmntent(MOUNTED, "a+")) == NULL)
		sysfatal("Can't open " MOUNTED);

	if (addmntent(mtab, &ment) == 1)
		sysfatal("Can't write mount entry");

	if (fchmod(fileno(mtab), 0644) == -1)
		sysfatal("Can't set perms on "MOUNTED);
		
	endmntent(mtab);

	if (unlink(MOUNTED"~") == -1)
		sysfatal("Can't remove "MOUNTED"~");
	
	free(mdev);
	mounted = 1;
}

int mountok(struct stat *st)
{
	if (!S_ISDIR(st->st_mode))
	{
		errno = ENOTDIR;
		return -1;
	}
	
	if (getuid() != 0 &&
	    (getuid() != st->st_uid ||
	     (st->st_mode & S_IRWXU) != S_IRWXU)
	   )
	{
		errno = EPERM;
		return -1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	struct stat st;
	int kid, i, ch;
	int p1[2], p2[2];
	char **args;
	char b1[10], b2[10];
	int async = 0;
	
	progname = argv[0];

	if (geteuid() != 0 || getegid() != getgid())
	{
		fprintf(stderr, "%s: must be installed suid root\n",
			progname);
		exit(1);
	}

	for(i = 0; (ch = getopt(argc, argv, "a")) != EOF;)
		switch(ch)
		{
		case 'a':
			async = 1;
			break;
		default:
			i++;
		}

	argc -= optind;
	argv = &argv[optind];
	
	if (i != 0 || argc < 2)
	{
		fprintf(stderr,
			"Usage: %s [-a] program mount-point [args...]\n",
			progname);
		exit(1);
	}

	if (async && getuid() != 0)
	{
		fprintf(stderr, "%s: async (-a) ignored: you can't unmount if you arn't root\n",
			progname);
		async = 0;
	}
	
	if (stat(argv[1], &st) == -1)
		sysfatal("Stat of %s failed", argv[1]);

	if (mountok(&st) == -1)
		sysfatal("can't mount on %s", argv[1]);

	if (pipe(p1) == -1 || pipe(p2) == -1)
		sysfatal("can't create pipe");

	for(i = 1; i < NSIG; i++)
	{
		struct sigaction sa;
		sigset_t ss;

		sigemptyset(&ss);
		
		sa.sa_handler = sighandler;
		sa.sa_flags = SA_NOMASK|SA_RESTART;
		sa.sa_mask = ss;
		
		if (i != SIGKILL && i != SIGSTOP)
			sigaction(i, &sa, NULL);
	}

	switch(kid = fork())
	{
	case -1:		/* fail */
		sysfatal("fork failed");

	case 0:			/* child */
		break;

	default:		/* parent */
		close(p1[1]);
		close(p2[0]);

		mounted = 0;

		if (checkfs(p1[0], p2[1]))
			domount(p1[0], p2[1], argv[1], argv[0], 
				&argv[2], argc-2, kid);
		else
			fprintf(stderr,
				"%s: filesystem failed to respond correctly.\n",
				progname);

		while (!async)
		{
			unsigned wstat;
			
			if (wait4(kid, &wstat, 0, NULL) == -1)
				fprintf(stderr, "%s: wait for pid %d failed: %s\n",
					progname, kid, strerror(errno));
			else
				if (!WIFEXITED(wstat))
					continue;
			
			if (mounted && !doumount(mount_point) != 0)
				fprintf(stderr, "%s: Failed to umount filesystem\n",
					progname);
			break;
		}
		exit(0);
	}

	/* child */
	close(p1[0]);
	close(p2[1]);
	
	if (setreuid(-1, getuid()) == -1)
		sysfatal("Can't drop suid permissions");

	{
		static int isigs[] = { SIGKILL, SIGINT };
		for(i = 0; i < sizeof(isigs)/sizeof(*isigs); i++)
		{
			struct sigaction sa;
			sigset_t ss;

			sigemptyset(&ss);
		
			sa.sa_handler = SIG_IGN;
			sa.sa_flags = SA_NOMASK|SA_RESTART;
			sa.sa_mask = ss;
		
			sigaction(isigs[i], &sa, NULL);
		}
	}
	args = (char **)xmalloc(sizeof(char *)* (argc + 3));

	sprintf(b1, "-i%d", p2[0]);
	sprintf(b2, "-o%d", p1[1]);
	
	args[0] = argv[0];
	args[1] = b1;
	args[2] = b2;
	args[3] = "-m";
	args[4] = argv[1];

	for(i = 0; i <= argc-3; i++)
		args[i+5] = argv[i+2];
	args[i+5] = NULL;
	
	execvp(args[0], args);

	sysfatal("exec of %s failed", args[0]);
	exit(1);
}
