// -*- C++ -*-
// Communications class
//
// This class copes with all the communication to and from the
// kernel for a userfs filesystem.
// The implementation of the base class is purely synchronous.
//
// This program is distributed under the terms of the
// Free Software Foundation General Public Licence.
// Copyright Jeremy Fitzhardinge <jeremy@sw.oz.au> 1993

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>

#pragma implementation

#include "Comm.h"
#include "Filesystem.h"

class doreq: public DispatchFD
{
	int tokern;
	Filesystem &filesys;
	int deferred;
public:
	doreq(int tokern, Filesystem &filesys);

	int dispatch(int fd);
	void defer(int fd);
};

Comm::Comm(Filesystem &fs, unsigned int to, unsigned int from)
     :filesys(fs),tokern(to)
{
	fdlist = NULL;
	req = new doreq(to, fs);
	addDispatch(from, req);
	
	filesys.setcomm(this);
}

Comm::~Comm()
{
	close(tokern);
	tokern = (unsigned)-1;
	while(fdlist)
		delDispatch(fdlist);
}

static int
recv(int from, unsigned char *buf, size_t sz)
{
	int	rd = 0;
	int	ret;
	
	while(sz > 0)
	{
		ret = read(from, buf, sz);

		if (ret == -1)
		{
			if (errno == EINTR)
				continue;

			return -1;
		}

		if (ret == 0)
			break;
				
		sz -= ret;
		buf += ret;
		rd += ret;
	}

	return rd;
}

static int
send(int tofd, const unsigned char *buf, size_t sz)
{
	int	wr = 0;
	int	ret;
	
	while(sz > 0)
	{
		ret = write(tofd, buf, sz);

		if (ret == -1)
		{
			if (errno == EINTR)
				continue;
			return -1;
		}

		if (ret == 0)
		{
			fprintf(stderr, "send write wrote 0 bytes\n");
			break;
		}

		sz -= ret;
		buf += ret;
		wr += ret;
	}

	return wr;
}

class copy : public DispatchFD
{
	int to;
public:
	copy(int to);
	dispatch(int fd);
};

copy::copy(int to)
     :to(to)
{
}

int
copy::dispatch(int from)
{
	int rd;
	char buf[1024];

 again:
	while((rd = read(from, buf, sizeof(buf))) > 0)
	{
		char *wrp = buf;

		while(rd > 0)
		{
			int wr = write(to, wrp, rd);

			if (wr == -1 && errno == EINTR)
				continue;
			
			if (wr == -1 || wr == 0)
				return -1;
			wrp += wr;
			rd -= wr;
		}
	}
	
	if (rd == -1)
	{
		if (errno == EINTR)
			goto again;
		return -1;
	}
	
	return 0;
}	

struct disp_fd
{
	disp_fd *next, *prev;
	int fd;
	DispatchFD *dfd;

	~disp_fd();
};

disp_fd::~disp_fd()
{
	::close(fd);
	delete dfd;
}

DispatchFD::~DispatchFD()
{
}

disp_fd *
Comm::addDispatch(int fd, DispatchFD *dfd)
{
	disp_fd *fdl = new disp_fd;

	fdl->prev = NULL;
	fdl->next = fdlist;
	if (fdl->next != NULL)
		fdl->next->prev = fdl;
	fdlist = fdl;

	fdl->fd = fd;
	fdl->dfd = dfd;

	return fdl;
}

void
Comm::delDispatch(disp_fd *fdl)
{
	if (fdl->dfd == req)
		req = NULL;

	if (fdl->next != NULL)
		fdl->next->prev = fdl->prev;
	if (fdl->prev != NULL)
		fdl->prev->next = fdl->next;
	else
		fdlist = fdl->next;

	::close(fdl->fd);
	fdl->next = fdl->prev = NULL;

	delete fdl;
}

// Wait on input filedescriptors for something to happen.
// Things which want to know about file descriptors (for read only)
// can register themselves.  When select() returns, each
// matching dispatch routine is called so it can do what it likes.
// If a dispatch routine returns 0, it means it wants to be removed
// from the list; -1 indicates error and 1 means OK.  If there
// are no routines on the list then Comm::Run returns.
int
Comm::Run()
{
	int ret;
	
	while(fdlist != NULL)
	{
		fd_set set;
		FD_ZERO(&set);

		disp_fd *dfd;

		for(dfd = fdlist; dfd != NULL; dfd = dfd->next)
			FD_SET(dfd->fd, &set);
		
		ret = select(NFDBITS, &set, NULL, NULL, NULL);

		if (ret == -1)
		{
			perror("select failed");
			if (errno == EINTR)
				continue;
			else
				return ret;
		}
		if (ret == 0)
		{
			fprintf(stderr, "Comm::Run ret == 0\n");
			continue;
		}
		
		disp_fd *next;
		for(dfd = fdlist; dfd != NULL; dfd = next)
		{
			next = dfd->next;
			if (!FD_ISSET(dfd->fd, &set))
				continue;

			ret = dfd->dfd->dispatch(dfd->fd);

			if (ret == 1)
				continue;
			delDispatch(dfd);
			if (ret != 0)
				return ret;
		}
	}

	return 0;
}

// Close filedescriptors we really don't need
void
Comm::CloseFDs(int all)
{
	disp_fd *fdlp, *next;
		
	for(fdlp = fdlist; fdlp != NULL; fdlp = next)
	{
		next = fdlp->next;
			
		if (all || fdlp->dfd != req)
			delDispatch(fdlp);
	}
}

// This can be (indirectly) called by filesystem methods so they
// can defer replying until later, allowing the filesystem to be
// used by other processes in the meantime.  This would happen,
// for example, if a read takes a long time to generate and the
// filesystem should be available for directory scans and name
// lookups.
// The function uses fork(); the parent returns so the main
// filesystem process can keep accepting requests.  The child
// works away and sends the reply when its ready.  This goes
// through a pipe from the child to the parent for copying
// to the kernel in one piece.  Everything is set up so the
// Comm class in the child knows what to do, and exits when
// done.
int
Comm::DeferRepl()
{
	int p[2];
	int pid;
	
	if (pipe(p) == -1)
	{
		perror("Comm::DeferRepl pipe() failed");
		return -1;
	}
	
	switch(pid = fork())
	{
	case -1:
		perror("Comm::DeferRepl fork() failed");
		return -1;
	case 0:
		CloseFDs(0);
		req->defer(p[1]);
		::close(p[0]);
		break;
	default:
		::close(p[1]);
		addDispatch(p[0], new copy(tokern));
		break;
	}

	return pid;
}

doreq::doreq(int tokern, Filesystem &fs)
      : tokern(tokern), filesys(fs)
{
	deferred = 0;
}

void
doreq::defer(int fd)
{
	::close(tokern);
	tokern = fd;
	deferred = 1;
}

// Process a request from the kernel
// Returns 0 on EOF, -1 on error and 1 for OK
doreq::dispatch(int fd)
{
	up_preamble pre;
	size_t presize = sizeof_up_preamble(&pre);
	Uchar hbuf[256];
	Uchar buf[4096];
	int ret;

	ret = recv(fd, hbuf, presize);
	if (ret == 0)
		return 0;
	
	if (ret != presize)
	{
		fprintf(stderr, "doreq::dispatch failed to get whole header (%d wanted, %d got)\n",
			presize, ret);
		return -1;
	}

	upp_repl repl;
	size_t replsize = sizeof_upp_repl(&repl);

	assert(sizeof(hbuf) > replsize);
		
	decode_up_preamble(&pre, hbuf);
		
	if (pre.size != 0 && (ret = recv(fd, buf, pre.size)) != pre.size)
	{
		fprintf(stderr, "doreq::dispatch: failed to get whole body (%d wanted, %d got)\n",
			pre.size, ret);
		return -1;
	}
	
	ret = filesys.DoOp(pre, repl, buf);

	if (ret == -1 && !deferred)
		return 1;

	if (!ret)
		return -1;

	if (repl.errno != 0)
	{
#ifdef DEBUG
		fprintf(stderr, "Request %d failing with %d %s\n",
			pre.op, repl.errno, strerror(repl.errno));
#endif
	}
		
	encode_upp_repl(&repl, hbuf);

	if (send(tokern, hbuf, replsize) != replsize)
		return -1;
		
	if (repl.size != 0 && send(tokern, buf, repl.size) != repl.size)
		return -1;

	return deferred ? 0 : 1;
}
