// -*- C++ -*-

// Generic directory inode by Jeremy Fitzhardinge
// This program is distributed under the terms of the
// Free Software Foundation General Public Licence.

// A directory

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <assert.h>
#include <sys/time.h>

#pragma implementation

#include "Filesystem.h"
#include "SimpleInode.h"
#include "DirInode.h"
#include "String.h"
#include "dbg.h"

DirInode::DirInode(Filesystem &fs, Handle h, Inode *p, int mm)
	:SimpleInode(fs, h), max_multird(mm)
{
	LOG("DirInode::DirInode");

	mode = S_IFDIR;
	
	parent = p == NULL ? this : p;

	// The top of a filesystem has an extra magic link
	if (p == NULL)
		nlink++;
	
	assert(entries.empty());
	link(".", this);
	link("..", parent);

	size = entries.length();
}

int
DirInode::link(const String name, Inode *ino)
{
	if (lookup(name) != NULL)
		return EEXIST;
	
	entries.append(*new DirEntry(name, ino));
	ino->incref();
	
	size++;

	return 0;
}

Pix
DirInode::pixlookup(const String name)
{
	for (Pix p = entries.first(); p != NULL; entries.next(p))
	{
		DirEntry *de = &entries(p);

		if (de->name == name)
			return p;
	}

	return NULL;
}

DirEntry *
DirInode::lookup(const String name)
{
	Pix p = pixlookup(name);

	if (p != NULL)
		return &entries(p);
	else
		return NULL;
}

DirEntry *
DirInode::scan(Pix &p)
{
	if (p == NULL)
		p = entries.first();

	if (p == NULL)
		return NULL;
	
	entries.next(p);
	
	return &entries(p);
}

DirEntry *
DirInode::scan(int &ind)
{
	if (ind >= entries.length())
		return NULL;

	int i;
	Pix p;
	for(p = entries.first(), i = 0;
	    p != NULL, i < ind;
	    entries.next(p))
		;
	
	if (p == NULL)
		return NULL;

	ind++;
	return &entries(p);
}

int
DirInode::unlink(const String name)
{
	update();

	Pix p = pixlookup(name);

	if (p == NULL)
		return ENOENT;

	DirEntry *de = &entries(p);
	de->geti()->decref();
	entries.del(p);

	size--;
	
	return 0;
}

// Implementations of userfs operations
int
DirInode::do_unlink(const up_preamble &, upp_repl &,
		    const upp_unlink_s &arg)
{
	return unlink(String(arg.name.elems, arg.name.nelem));
}

int
DirInode::do_link(const up_preamble &, upp_repl &,
		  const upp_link_s &args)
{
	update();
	
	Inode *ino = findino(args.ofile.handle);

	if (!ino)
		return EINVAL;

	return link(String(args.name.elems, args.name.nelem), ino);
}

int
DirInode::do_rename(const up_preamble &, upp_repl &,
		    const upp_rename_s &arg)
{
	Inode *ndir = findino(arg.ndir.handle);

	if (ndir == NULL || !S_ISDIR(ndir->mode))
		return ENOTDIR;

	update();

	String oname(arg.oname.elems, arg.oname.nelem);
	
	DirEntry *fde = lookup(oname);

	if (!fde)
		return ENOENT;
	
	Inode *file = fde->geti();
	
	if (file == NULL)
		return EINVAL;

	int ret = ndir->link(arg.nname.elems, arg.nname.nelem, file);
	if (ret)
		return ret;
	ret = unlink(oname);

	ndir->mtime = mtime = time(0);
	
	return ret;
}

int
DirInode::do_lookup(const up_preamble &pre, upp_repl &rep,
		    const upp_lookup_s &arg, upp_lookup_r &res)
{
	update();

	DirEntry *de = lookup(String(arg.name.elems, arg.name.nelem));

	atime = time(0);
	
	if (de != NULL)
	{
		res.handle.handle = de->geti()->gethandle();
		return 0;
	}
	return ENOENT;
}

int
DirInode::do_readdir(const up_preamble &pre, upp_repl &repl, 
		     const upp_readdir_s &arg, upp_readdir_r &res)
{
	LOG("readdir");

	update();
	
	res.off=1;
	if(arg.off >= size)
	{
		res.off=0;
		res.name.nelem=0;
		return 0;
	}

	atime = time(0);
	
	Pix pix=entries.first();
	
	for(int i=0; i < arg.off; i++)
		entries.next(pix);

	DirEntry *p = &(entries(pix));

	res.name.nelem = strlen(p->name);
	res.name.alloc(res.name.nelem+1);
	strcpy(res.name.elems, p->name);
	res.file.handle = p->ino->gethandle();
	
	DB(printf("Readdir:'%s',%d,%d\n",res.name.elems,res.name.nelem,res.file.handle));
	return 0;
}

int
DirInode::do_read(const up_preamble &, upp_repl &,
		  const upp_read_s &, upp_read_r &)
{
	return EISDIR;
}

int
DirInode::do_write(const up_preamble &, upp_repl &,
		   const upp_write_s &, upp_write_r &)
{
	return EISDIR;
}

int
DirInode::do_multireaddir(const up_preamble &pre, upp_repl &repl,
			  const upp_multireaddir_s &arg,
			  upp_multireaddir_r &res)
{
	off_t off = arg.off;

	update();
	
	LOG("multireaddir");

	if (arg.off >= size)
	{
		res.nelem = 0;
		return 0;
	}
	
	atime = time(0);
	
	Pix pix=entries.first();
	
	for(int i=0; i < arg.off; i++)
		entries.next(pix);

	upp_readdir_r *ents = new upp_readdir_r[max_multird];
	res.elems = ents;
	res.nelem = 0;
	
	for(i = 0; i < max_multird && off < size &&
	    sizeof_upp_multireaddir_r(&res) < USERFS_MAX_XFER;
	    i++, off++, entries.next(pix))
	{
		DirEntry *p = &(entries(pix));

		ents[i].off = 1;
		ents[i].name.nelem = p->name.length();
		ents[i].name.alloc(ents[i].name.nelem+1);
		strcpy(ents[i].name.elems, (const char *)p->name);
		ents[i].file.handle = p->ino->gethandle();
		DB(printf("Multireaddir:%d '%s',nlen=%d,ino=%d,off=%d\n",
			  i, ents[i].name.elems,
			  ents[i].name.nelem,
			  ents[i].file.handle, off));
		res.nelem = i+1;
	}
	res.nelem = i;
	
	while(sizeof_upp_multireaddir_r(&res) >= USERFS_MAX_XFER)
		res.nelem--;
	
	return 0;
}

DirEntry::DirEntry(const String name, Inode *i)
	: name(name), ino(i)
{
}

DirEntry::~DirEntry()
{
}
